package com.qmth.themis.task.quartz.service.impl;

import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.qmth.themis.business.entity.TEExamActivity;
import com.qmth.themis.business.entity.TEExamStudent;
import com.qmth.themis.business.entity.TOeExamRecord;
import com.qmth.themis.business.enums.ExamRecordStatusEnum;
import com.qmth.themis.business.enums.FinishTypeEnum;
import com.qmth.themis.business.service.TEExamActivityService;
import com.qmth.themis.business.service.TEExamStudentService;
import com.qmth.themis.business.service.TOeExamRecordService;
import com.qmth.themis.mq.service.MqDtoService;
import com.qmth.themis.task.quartz.service.QuartzLogicService;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import javax.annotation.Resource;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Objects;

/**
 * @Description: quartz逻辑 impl
 * @Param:
 * @return:
 * @Author: wangliang
 * @Date: 2020/7/31
 */
@Service
public class QuartzLogicServiceImpl implements QuartzLogicService {
    private final static Logger log = LoggerFactory.getLogger(QuartzLogicServiceImpl.class);

    @Resource
    TEExamActivityService teExamActivityService;

    @Resource
    TEExamStudentService teExamStudentService;

    @Resource
    TOeExamRecordService tOeExamRecordService;

    @Resource
    MqDtoService mqDtoService;

    /**
     * 考试场次quartz逻辑
     *
     * @param key
     */
    @Override
    @Transactional
    public void execExamActivityJobLogic(String key) {
        log.info("key:{}", key);
        String[] strings = key.split("\\.");
        QueryWrapper<TEExamActivity> teExamActivityQueryWrapper = new QueryWrapper<>();
        teExamActivityQueryWrapper.lambda().eq(TEExamActivity::getCode, strings[1]);
        TEExamActivity teExamActivity = teExamActivityService.getOne(teExamActivityQueryWrapper);
        if (Objects.nonNull(teExamActivity)) {
            //获取该考试批次下所有未交卷的考生的考试记录
            QueryWrapper<TOeExamRecord> tOeExamRecordQueryWrapper = new QueryWrapper<>();
            tOeExamRecordQueryWrapper.lambda().eq(TOeExamRecord::getExamActivityId, teExamActivity.getId()).ne(TOeExamRecord::getStatus, ExamRecordStatusEnum.finished.ordinal());
            List<TOeExamRecord> tOeExamRecordList = tOeExamRecordService.list(tOeExamRecordQueryWrapper);
            List<Long> examStudentIdList = null;
            if (Objects.nonNull(tOeExamRecordList) && tOeExamRecordList.size() > 0) {
                examStudentIdList = new ArrayList<>();
            }
            List<Long> finalExamStudentIdList = examStudentIdList;
            tOeExamRecordList.forEach(s -> {
                s.setStatus(ExamRecordStatusEnum.finished);
                s.setFinishTime(new Date());
                s.setFinishType(FinishTypeEnum.interrupt);
                finalExamStudentIdList.add(s.getExamStudentId());
            });
            tOeExamRecordService.updateBatchById(tOeExamRecordList);

            if (Objects.nonNull(examStudentIdList) && examStudentIdList.size() > 0) {
                //获取该考试批次下所有考生，考试次数减1
                QueryWrapper<TEExamStudent> teExamStudentQueryWrapper = new QueryWrapper<>();
                teExamStudentQueryWrapper.lambda().in(TEExamStudent::getId, examStudentIdList);
                List<TEExamStudent> teExamStudentList = teExamStudentService.list(teExamStudentQueryWrapper);
                teExamStudentList.forEach(s -> {
                    int count = Objects.isNull(s.getLeftExamCount()) ? 0 : s.getLeftExamCount();
                    count--;
                    s.setLeftExamCount(count < 0 ? 0 : count);
                    //加入踢下线mq
                });
                teExamStudentService.updateBatchById(teExamStudentList);
            }
            //todo 未完待续,需要加入交卷逻辑
        } else {
            log.info("考试场次:{}已删除", key);
        }
    }
}
